/*******************************************************************
 *             Copyright (c) 2001 by TAOS Networks, Inc.
 *                       All rights reserved.
 *  
 *  This file is proprietary and confidential to TAOS Networks, Inc.
 *  No part of this file may be reproduced, stored, transmitted,
 *  disclosed or used in any form or by any means other than as
 *  expressly provided by the written permission from Jianhui Tao
 * 
 ******************************************************************/

#include <stdio.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <semaphore.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "tmodule.h"
#include "hpnsUtil.h"

#define tfree(x) { if ( x ) { free(x); x = 0; } }

#ifdef __ANDROIDOS__
  #include <jni.h>
  extern jclass g_classAdp;
  extern JavaVM* g_JavaVm;
  extern JNIEnv* g_jniEnv;
#endif


void      *taosProcessQueue(void *param);

char *taosDisplayModuleStatus(int moduleNum) {
  static char  status[256];
  int i;

  status[0] = 0;
  
  for (i=1; i < moduleNum; ++i)
    if ( moduleObj[i].thread != 0 )
      sprintf(status+strlen(status), "%s ", moduleObj[i].name);
  
  if ( status[0] == 0 ) 
    sprintf(status, "all module is down");
  else
    sprintf(status, " is(are) up");

  return status;
}

/*
 * allocate the storage
 * the module is used to store the corresponding message with a queue
 * the queue stat is reflected on the module's semaphore: fullSem and emptySem.
 */
taosInitModule (module_t *pMod)
{
  pthread_attr_t attr;
  
  if ( pthread_mutex_init ( &pMod->queueMutex, 0) < 0 ) {
    printf("ERROR: init %s queueMutex failed, reason:%s\n", pMod->name, strerror(errno));
    taosCleanUpModule(pMod);
    return -1;
  }
  
  if ( pthread_mutex_init ( &pMod->stmMutex, 0) < 0 ) {
    printf("ERROR: init %s stmMutex failed, reason:%s\n", pMod->name, strerror(errno));
    taosCleanUpModule(pMod);
    return -1;
  }

  if ( sem_init (&pMod->emptySem, 0, pMod->queueSize) != 0 ) {
    printf("ERROR: init %s empty semaphore failed, reason:%s\n", pMod->name, strerror(errno));
    taosCleanUpModule(pMod);
    return -1;
  }

  if ( sem_init (&pMod->fullSem, 0, 0) != 0 )  {
    printf("ERROR: init %s full semaphore failed, reason:%s\n", pMod->name, strerror(errno));
    taosCleanUpModule(pMod);
    return -1;
  }

  if ( ( pMod->queue = (msg_t *)malloc(pMod->queueSize*sizeof(msg_t)) ) == 0 ) {
    printf("ERROR: %s no enough memory, reason:%s\n", pMod->name, strerror(errno));
    taosCleanUpModule(pMod);
    return -1;
  }
	   
  memset(pMod->queue, 0, pMod->queueSize * sizeof(msg_t) );
  pMod->fullSlot = 0;
  pMod->emptySlot = 0;
  pMod->msgCount = 0 ;

  pthread_attr_init ( &attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
  
  if ( pthread_create(&pMod->thread, &attr, taosProcessQueue, (void *)pMod) != 0 ) {
    printf("ERROR: %s failed to create thread, reason:%s\n", pMod->name, strerror(errno));
    taosCleanUpModule(pMod);
    return -1;
  }

  if ( pMod->init )
    return (*(pMod->init))();

  return 0;
}

void *taosProcessQueue(void *param) {
  msg_t    msg;
  module_t *pMod = (module_t *)param;
  int      oldType;
  sigset_t intmask;

#ifdef __ANDROIDOS__
	JNIEnv* Env = NULL;
	(*g_JavaVm)->AttachCurrentThread(g_JavaVm, &Env, NULL);
#endif
  
  //pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, &oldType);  
  
  signal( SIGINT, SIG_IGN);

  pMod->threadId = gettid() ;

  while ( 1 ) {
    while ( sem_wait ( &pMod->fullSem ) != 0 )
    {
        if( errno != EINTR);
        {
           printf("ERROR: wait %s fullSem failed, reason:%s\n", pMod->name, strerror(errno));
           break;
        } 
    }

    if ( pthread_mutex_lock ( &pMod->queueMutex) != 0 ) 
      printf("ERROR: lock %s queueMutex failed, reason:%s\n", pMod->name, strerror(errno));
    
    msg = pMod->queue[pMod->fullSlot];
    memset( &(pMod->queue[pMod->fullSlot]), 0, sizeof(msg_t));
    pMod->fullSlot = ( pMod->fullSlot + 1 ) % pMod->queueSize;
    pMod->msgCount-- ;
    
    if ( pthread_mutex_unlock ( &pMod->queueMutex) != 0 )
      printf("ERROR: unlock %s queueMutex failed, reason:%s\n", pMod->name, strerror(errno));

    while  ( sem_post (&pMod->emptySem) != 0 )
    {
        if ( errno != EINTR )
        {
            printf("ERROR: post %s emptySem failed, reason:%s\n", pMod->name, strerror(errno));
            break;
        }
     }

    /* process the message */
    if ( msg.cid<0 || msg.cid >= maxCid ) {
      printf("ERROR: cid:%d is out of range, msg is discarded\n", msg.cid);
      continue;
    }

    if ( pthread_mutex_lock ( &(pMod->stmMutex)) != 0 )
      printf("ERROR: lock %s stmMutex failed, reason:%s\n", pMod->name, strerror(errno));
		      
    (*(pMod->processMsg)) ( &msg );

    if (msg.tid > 0 && msg.msg)
        tfree(msg.msg);
    
    if ( pthread_mutex_unlock ( &(pMod->stmMutex)) != 0 )
      printf("ERROR: unlock %s stmMutex failed, reason:%s\n", pMod->name, strerror(errno));

  }

#ifdef __ANDROIDOS__
	(*g_JavaVm)->DetachCurrentThread(g_JavaVm);
#endif
}

int taosSendMsgToModule(module_t *pMod, int cid, int mid, int tid, char *msg) {

  if (pMod->msgCount > pMod->queueProtectSize) {
        if (gettid() == pMod->threadId) {
            nprintf("tmodule error that message count: %d in pMod: %s is exceed protect size, just discard the self msg, cid: %d, mid: %d", pMod->msgCount, pMod->name, cid, mid) ;
            return ;
        }
  }
  
  while ( sem_wait( &pMod->emptySem ) != 0 )
  {
      if ( errno != EINTR ) 
      {
        printf("ERROR: wait %s emptySem failed, reason:%s\n", pMod->name, strerror(errno));
        break;
      }
  }
  if ( pthread_mutex_lock ( &pMod->queueMutex ) != 0 ) 
    printf("ERROR: lock %s queueMutex failed, reason:%s\n", pMod->name, strerror(errno));
  
  pMod->queue[pMod->emptySlot].cid = cid;
  pMod->queue[pMod->emptySlot].mid = mid;
  pMod->queue[pMod->emptySlot].tid = tid;
  if (tid > 0 && msg) {
    pMod->queue[pMod->emptySlot].msg = (char*)malloc(tid);
    memcpy(pMod->queue[pMod->emptySlot].msg, msg, tid);
  } else {
    pMod->queue[pMod->emptySlot].msg = msg;
  }
  pMod->emptySlot = (pMod->emptySlot+1) % pMod->queueSize;
  pMod->msgCount++ ;
  if ( pthread_mutex_unlock( &pMod->queueMutex ) != 0 ) 
    printf("ERROR: unlock %s queueMutex failed, reason:%s\n", pMod->name, strerror(errno));
  
  while ( sem_post ( &pMod->fullSem) != 0 )
  { 
      if ( errno != EINTR )
      {
        printf("ERROR: post %s fullSem failed, reason:%s\n", pMod->name, strerror(errno) );
        break;
      }
   }
  return 0;
}

void taosCleanUpModule(module_t *pMod ) {
  int i;

  if ( pMod->cleanUp ) 
    pMod->cleanUp();

  //if ( pMod->thread ) 
    //pthread_cancel(pMod->thread);
  pMod->thread = 0;  
 
  sem_destroy( &pMod->emptySem);
  sem_destroy( &pMod->fullSem );
  pthread_mutex_destroy( &pMod->queueMutex );
  pthread_mutex_destroy( &pMod->stmMutex );

  for ( i=0; i<pMod->queueSize; ++i) {
    tfree(pMod->queue[i].msg);
  }

  tfree(pMod->queue);

  //memset(pMod, 0, sizeof(module_t));

}
